/* Copyright 2003 Heikki Orsila <heikki.orsila@ee.tut.fi>
   Copyright 2003 Harry Sintonen
*/

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <ctype.h>
#ifndef PATH_MAX
#define PATH_MAX 1024
#endif

#include "amiga-shell.h"

#include "uade-os.h"
#include "../osdep/strl.c"

#include <exec/types.h>
#include <dos/dos.h>
#include <dos/dosextens.h>
#include <dos/dosasl.h>
#include <proto/dos.h>


void as_setup(void (*interrupt)(void));
void as_enqueue_song(char *file);
void as_get_next(char *dst, int maxsize);
int as_song_list_empty(void);


struct songnode_t {
  struct songnode_t *next;
  char filename[PATH_MAX];
};

static struct songnode_t *SongList = 0;

static void (*as_interrupt)(void) = 0;

/* catch CTRL-C and cause REBOOT */
static void as_sigint_handler(int signum) {
  signum = signum;
  if (as_interrupt) {
    as_interrupt();
  } else {
    fprintf(stderr, "uade: non-initialized sigint interrupt\n");
  }
}

static int as_cleanup_installed = 0;
static BPTR as_stdin = NULL;
static BPTR as_origdir = NULL;
static int as_stdin_does_cooked = 0;
static int as_stdin_is_cooked = 0;

static
void as_cleanup(void) {
  if (as_stdin) {
    if (as_stdin_is_cooked) {
      as_stdin_is_cooked = 0;
      /* back to con mode */
      SetMode(as_stdin, 0);
    }
    as_stdin = NULL;
  }
  if (as_origdir)
  {
    BPTR lock;
    lock = CurrentDir(as_origdir);
    if (lock != as_origdir)
    {
      UnLock(lock);
    }
    as_origdir = NULL;
  }
}

void as_setup(void (*interrupt)(void)) {
  if (!as_cleanup_installed) {
    as_cleanup_installed = 1;
    if (atexit(as_cleanup) < 0) {
      uade_exit(-1);
    }
  }

  as_stdin = Input();

  if (!as_stdin_does_cooked) {
    /* only do console interaction if stdin is interactive */
    if (IsInteractive(as_stdin)) {
      /* enter RAW 'cooked' mode */
      if (!SetMode(as_stdin, 1)) {
        printf("as_setup: can't put console to cooked raw mode!\n");
        uade_exit(-1);
      }
      /* back to con mode */
      SetMode(as_stdin, 0);
      /* yeah, it does cooked mode */
      as_stdin_does_cooked = 1;
    }
  }

  as_origdir = CurrentDir(NULL);
  (void) CurrentDir(as_origdir);

  as_interrupt = interrupt;
  /* create CTRL-C break for cmdline tool */
  uade_create_signalhandler(as_sigint_handler, UADE_SIGINT);
}


static
void enqueue_song(char *file) {
  struct songnode_t *s = SongList;
  struct songnode_t *new;

  /* allow queuing of zero length string */
  if (uade_is_regular_file(file) || strlen(file) == 0) {
    new = malloc(sizeof(struct songnode_t));
    new->next = 0;
    strlcpy(new->filename, file, sizeof(new->filename));
    if (s) {
      while(s->next) s = s->next;
      s->next = new;
    } else {
      SongList = new;
    }
  }
}


struct anchordata {
  struct AnchorPath ap;
  char buf[1024];
};

void as_enqueue_song(char *file) {
  char _data[sizeof(struct anchordata) + 3];
  struct anchordata *data = (void *) (((ULONG) _data + 3) & ~3);
  LONG err, count = 0;

  data->ap.ap_BreakBits  = SIGBREAKF_CTRL_C;
  data->ap.ap_FoundBreak = 0;
  data->ap.ap_Flags      = 0;
  data->ap.ap_Strlen     = sizeof(data->buf);

  for (err = MatchFirst(file, &data->ap); !err; err = MatchNext(&data->ap)) {
    enqueue_song(data->ap.ap_Buf);
    count++;
  }

  /* If nothing matched, assume it's literal name */
  if (count == 0 && err == ERROR_NO_MORE_ENTRIES) {
    enqueue_song(file);
  }
}

int as_song_list_empty(void) {
  return SongList ? 0 : 1;
}

void as_get_next(char *dst, int maxsize) {
  char playername[PATH_MAX];
  char tempname[PATH_MAX];

  if (!SongList)
    return;

  strlcpy(dst, SongList->filename, maxsize);
  SongList = SongList->next;
}

static int interaction_prompt_done = 0;
static int cmd_buf_size = 0;
static char cmd_buf[16+PATH_MAX];
static char buf[PATH_MAX];

/* strip leading whitespace from cmd_buf */
static void strip_whitespace(void) {
  int i;
  i = 0;
  while (i < cmd_buf_size) {
    if (!isspace((int) cmd_buf[i])) {
      break;
    }
    i++;
  }
  memmove(cmd_buf, &cmd_buf[i], cmd_buf_size - i);
  cmd_buf_size -= i;
}

static int get_rest_of_line(char *buf, int maxlen) {
  int i;
  for (i=0; i<cmd_buf_size; i++) {
    if (cmd_buf[i] == '\n') {
      cmd_buf[i] = 0;
      strlcpy(buf, cmd_buf, maxlen);
      memmove(cmd_buf, &cmd_buf[i+1], cmd_buf_size - (i + 1));
      cmd_buf_size -= i + 1;
      return 1;
    }
  }
  return 0;
}

/* get a word from cmd_buf */
static int pop_word(char *buf, int maxlen) {
  int i;
  strip_whitespace();
  for (i=0; i<cmd_buf_size; i++) {
    if (isspace((int) cmd_buf[i])) {
      cmd_buf[i] = 0;
      strlcpy(buf, cmd_buf, maxlen);
      memmove(cmd_buf, &cmd_buf[i+1], cmd_buf_size - (i + 1));
      cmd_buf_size -= i + 1;
      return 1;
    }
  }
  return 0;
}

static int as_interaction(struct uade_command *cmd, int wait_for) {
  LONG c;

  if (!as_stdin_does_cooked) {
    goto no_interaction;
  }

  /* enter RAW 'cooked' mode */
  if (!as_stdin_is_cooked) {
    if (!SetMode(as_stdin, 1)) {
      goto no_interaction;
    }

    as_stdin_is_cooked = 1;

    /* flush possible pending chars in the buffer */
    Flush(as_stdin);
  }

  if (!interaction_prompt_done)
  {
    printf("uade interactive mode, h for help\n");
    interaction_prompt_done = 1;
  }

  /* timeout 0 == poll */
  if (WaitForChar(as_stdin, 0)) {

    c = FGetC(as_stdin);

    switch (c)
    {
      case -1:
	goto no_interaction;

      case 0x03: /* CTRL-C */
        as_sigint_handler(1);
        break;

      case 0x9b:
	{
	  /* read esc sequence (fkeys, other special keys) */
	  char seq[12];
	  int i = 0;
	  do {
	    c = FGetC(as_stdin);
	    seq[i++] = c;
	  } while (c != -1 && c != '~' && i < (sizeof(seq) - 1));
	  seq[i] = '\0';
	  break;
	}
      case '\n':
      case '\r':
	break;

      case 'n':
	cmd->type = UADE_SONG_END;
	return 1;

      case 'm':
	cmd->type = UADE_REBOOT;
	return 1;

      case 'f':
        cmd->type = UADE_TOGGLE_LED;
        return 1;

      case 'h':
        printf("<number>\tGo to subsong <number>\n");
        printf("f\t\tToggle LED Filter\n");
        printf("h\t\tDisplay help\n");
        printf("m\t\tNext song\n");
        printf("n\t\tNext (sub)song\n");
        printf("q\t\tQuit uade\n");
        printf("!\t\tplaylist command\n");
        printf("  cd <dir>\tSwitch to directory <dir>\n");
        printf("  flush\t\tFlush playlist\n");
        printf("  play <pattern>\tEnqueue songs matching shell pattern\n");
        break;

      case 'q':
        uade_exit(0);
        return 1;

      case '!':
        {
          int ret = 0;
          /* back to con mode */
          SetMode(as_stdin, 0);
          PutStr("> "); Flush(Output());

          Flush(as_stdin);
          if (FGets(as_stdin, cmd_buf, sizeof(cmd_buf))) {
            cmd_buf_size = strlen(cmd_buf);
            if (pop_word(buf, sizeof(buf))) {
              strip_whitespace();
              if (!strcmp(buf, "cd")) {
                if (get_rest_of_line(buf, sizeof(buf))) {
                  BPTR lock;
                  lock = Lock(buf, ACCESS_READ);
                  if (lock) {
                    lock = CurrentDir(lock);
                    if (lock != as_origdir) {
                      UnLock(lock);
                    }
                  }
                  else {
                    printf("could not lock directory \"%s\"\n");
                  }
                }
              }
              else if (!strcmp(buf, "flush"))
              {
                while (!as_song_list_empty()) {
                 as_get_next(buf, sizeof(buf));
                }
              }
              else if (!strcmp(buf, "next"))
              {
                cmd->type = UADE_SONG_END;
                ret = 1;
              }
              else if (!strcmp(buf, "play"))
              {
                if (get_rest_of_line(buf, sizeof(buf))) {
                  as_enqueue_song(buf);
                }
              }
            }
          }
          /* and to interactive mode again */
          SetMode(as_stdin, 1);
          if (ret) {
            return ret;
          }
        }
        break;

      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
	cmd->type = UADE_SETSUBSONG;
	cmd->ret = malloc(sizeof(int));
	if (!cmd->ret) {
	  goto no_interaction;
	}
	*((int *) cmd->ret) = (int) (c - '0');
	return 1;

      default:
	fprintf(stderr, "uade: unknown command for interaction\n");
	break;
    }
  }

  return 0;

 no_interaction:
  cmd->type = UADE_NO_INTERACTION;
  return 1;
}

void as_functions(struct uade_local *local) {
  local->setup = as_setup;
  local->enqueue_song = as_enqueue_song;
  local->get_next = as_get_next;
  local->interaction = as_interaction;
  local->song_list_empty = as_song_list_empty;
};
