/* Copyright 2003 Heikki Orsila <heikki.orsila@ee.tut.fi>

TODO:
interactive mode: ls, command launch support, pwd, ...
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <dirent.h>
#include <glob.h>
#include <unistd.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <termios.h>
#include <errno.h>
#include <limits.h>
#ifndef PATH_MAX
#define PATH_MAX 1024
#endif
#include "uade-os.h"
#include "../osdep/strl.c"
#include "unix-shell.h"

#include "unix-shell-int.h"

#define INTERACTION_EOL 1
#define INTERACTION_TAB 2

static int interaction_setup = 1;
static int tab_state = 0;
static int tab_pos = 0;
static int prompt_shown = 0;
static int cmd_number = 1;
static int cmd_buf_size = 0;
static char cmd_buf[4096];
static char auto_pattern[PATH_MAX];

/* strip leading whitespace from cmd_buf */
static void strip_whitespace(void) {
  int i;
  i = 0;
  while (i < cmd_buf_size) {
    if (cmd_buf[i] == '\n') {
      prompt_shown = 0;
    }
    if (!isspace((int) cmd_buf[i])) {
      break;
    }
    i++;
  }
  memmove(cmd_buf, &cmd_buf[i], cmd_buf_size - i);
  cmd_buf_size -= i;
}

static void strip_end(char *buf) {
  int p = strlen(buf) - 1;
  if (p < 0) {
    return;
  }
  for (; p >= 0; p--) {
    if (!isspace(buf[p])) {
      break;
    }
    buf[p] = 0;
  }
}

static int get_rest_of_line(char *buf, int maxlen) {
  int i;
  for (i = 0; i < cmd_buf_size; i++) {
    if (cmd_buf[i] == '\n') {
      cmd_buf[i] = 0;
      strlcpy(buf, cmd_buf, maxlen);
      memmove(cmd_buf, &cmd_buf[i+1], cmd_buf_size - (i + 1));
      cmd_buf_size -= i + 1;
      return 1;
    }
  }
  return 0;
}

/* shifts command buffer left from position p by l characters */
static void shift_left(char *buf, int *s, int p, int l) {
  if (p < 0 || p >= *s) {
    fprintf (stderr, "uade: shift_left error: p = %d\n", p);
    return;
  }
  if (l < 0) {
    fprintf (stderr, "uade: shift_left error: l < 0\n");
    return;
  }
  if (l == 0) {
    return;
  }
  if ((p + l) >= *s) {
    *s = p;
    return;
  }
  memmove(&buf[p], &buf[p+l], *s - (p + l));
  *s -= l;
}

static void insert_to_buf(char *d, int *ds, int maxl, int p, char *s, int ss) {
  if (ss == 0 || maxl == 0) {
    return;
  }
  if (*ds < 0 || maxl < 0 || p < 0 || ss < 0) {
    fprintf(stderr, "uade: insert_to_buf: *ds = %d maxl = %d p = %d ss = %d\n", *ds, maxl, p, ss);
    return;
  }
  if (p >= maxl) {
    fprintf(stderr, "uade: insert_to_buf: p >= maxl\n");
    return;
  }
  if ((*ds + ss) > maxl) {
    fprintf(stderr, "uade: insert_to_buf: cmd_buf overflow\n");
    return;
  }
  if (p < *ds) {
    memmove(&d[p + ss], &d[p], (*ds - p));
  } else {
    *ds = p;
  }
  memcpy(&d[p], s, ss);
  *ds += ss;
}

/* get a word from cmd_buf */
static int pop_word(char *buf, int maxlen) {
  int i, n, left;
  strip_whitespace();
  for (i = 0; i < cmd_buf_size; i++) {
    if (isspace((int) cmd_buf[i])) {
      n = i;
      if (n >= maxlen) {
	n = maxlen - 1;
      }
      strncpy(buf, cmd_buf, n);
      buf[n] = 0;
      left = (cmd_buf[i] == '\n') ? i : (i + 1); /* don't shift '\n' away */
      shift_left(cmd_buf, &cmd_buf_size, 0, left);
      return 1;
    }
  }
  return 0;
}

static int scan_cmd_buf(void) {
  int i;
  strip_whitespace();
  for (i=0; i<cmd_buf_size; i++) {
    if (cmd_buf[i] == '\n') {
      return INTERACTION_EOL;
    } else if (cmd_buf[i] == '\t') {
      tab_pos = i;
      shift_left(cmd_buf, &cmd_buf_size, i, 1);
      return INTERACTION_TAB;
    }
  }
  return 0;
}

static int compselect(const struct dirent *e) {
  return !strncmp(e->d_name, auto_pattern, strlen(auto_pattern));
}

static void auto_completion(void) {
  int p = tab_pos - 1;
  int nchars;
  char pattern[PATH_MAX];
  char path[PATH_MAX];
  char addition[PATH_MAX];
  int nadd;
  char *slash;
  struct dirent **namelist;
  int ret;
  DIR *d;
  int i;

  if (p < 0) {
    return;
  }
  nchars = 0;
  while (p > 0 && !isspace(cmd_buf[p])) {
    nchars++;
    p--;
  }
  if (nchars == 0 || p == 0) {
    return;
  }
  p++;
  if (nchars >= ((int) sizeof(pattern))) {
    fprintf(stderr, "uade: unexpected buffer shortage\n");
    return;
  }
  memcpy(pattern, &cmd_buf[p], nchars);
  pattern[nchars] = 0;
  /* fprintf(stderr, "would complete %s\n", pattern); */

  slash = strrchr(pattern, (int) '/');
  if (slash) {
    int l = (int) (slash - pattern);
    int pattlen = strlen(pattern) + 1;
    memcpy(path, pattern, l + 1);
    path[l + 1] = 0;
    shift_left(pattern, &pattlen, 0, l + 1);
    /* fprintf(stderr, "path = %s pattern = %s\n", path, pattern); */
  } else {
    strlcpy(path, ".", sizeof(path));
  }

  /* find matching patterns */
  strlcpy(auto_pattern, pattern, sizeof(auto_pattern));
  ret = scandir(path, &namelist, compselect, alphasort);

  /* check pattern results */
  if (ret >= 0) {
    switch (ret) {
    case 0:
      /* no matching patterns. shit. */
      break;
    case 1:
      /* straight auto-completion */
      strlcat(path, namelist[0]->d_name, sizeof(path));
      d = opendir(path);
      nchars = strlen(namelist[0]->d_name);
      nadd = 0;
      for (p = strlen(pattern); p < nchars ; p++) {
	printf("%c", namelist[0]->d_name[p]);
	addition[nadd] = namelist[0]->d_name[p];
	nadd++;
      }
      if (d) {
	printf("/");
	addition[nadd] = '/';
	nadd++;
	closedir(d);
      } else {
	printf(" ");
	addition[nadd] = ' ';
	nadd++;
      }
      /* fprintf(stderr, "inserting '%s' to '%s'\n", addition, pattern); */
      insert_to_buf(cmd_buf, &cmd_buf_size, sizeof(cmd_buf), tab_pos, addition, nadd);
      fflush(stdout);
      tab_state = 0;
      break;
    default:
      /* print all alternatives (more than one) */
      printf("\n");
      for (i = 0; i < ret; i++) {
	printf("%s\n", namelist[i]->d_name);
	free(namelist[i]);
      }
      printf("uade %d> ", cmd_number);
      for (i = 0; i < tab_pos ; i++) {
	printf("%c", cmd_buf[i]);
      }
      fflush(stdout);
      break;
    }
    free(namelist);
  }
}

static int from_buf(char *c, int i) {
  if (i >= cmd_buf_size) {
    return 0;
  }
  *c = cmd_buf[i];
  return 1;
}

static void clean_up_parts(char **buf, int n) {
  int i;
  if (buf) {
    for (i = 0; i < n; i++) {
      if (buf[i]) {
	free (buf[i]);
      }
    }
    free (buf);
  }
}

/* divides the command line into parts separated with white space. is aware of
   '"' characters. */
static int divide_line(char ***buf) {
  char token[PATH_MAX];
  int i, j, n;
  char c, nc;
  int brace;
  int maxn;
  maxn = 1;
  *buf = malloc(maxn * sizeof(char *));
  if (!(*buf)) {
    fprintf (stderr, "uade: out of memory for command line parsing\n");
    return 0;
  }
  n = 0;
  i = 0;
  while (1) {
    if (!from_buf(&c, i)) {
      break;
    }
    if (isspace(c)) {
      if (c == '\n') {
	break;
      }
      i++;
      continue;
    }

    brace = 0;
    j = 0;
    while (1) {
      if (!from_buf(&c, i)) {
	break;
      }
      if (c == '\"') {
	if (brace) {
	  if (from_buf(&nc, i + 1)) {
	    if (isspace(nc)) {
	      i += 2;
	      break;
	    } else {
	      goto copy;
	    }
	  } else {
	    i++;
	    break;
	  }
	}
	brace = 1;
	i++;
	continue;

      } else if (c == '\\') {
	if (from_buf(&nc, i + 1)) {
	  c = nc;
	  i++; /* skip '\\' */
	}
      } else if (isspace(c)) {
	if (!brace) {
	  i++;
	  break;
	}
      }

    copy:
      if (j >= ((int) sizeof(token))) {
	fprintf(stderr, "uade: command editing buffer exceeded\n");
	break;
      }
      token[j++] = c;
      i++;
    }
    token[j] = 0;

    if (n >= maxn) {
      char **newbuf;
      maxn *= 2;
      newbuf = realloc(*buf, maxn * sizeof(char *));
      if (!newbuf) {
	goto line_clean_up;
      }
      *buf = newbuf;
    }
    (*buf)[n] = strdup(token);
    if (!(*buf)[n]) {
      goto line_clean_up;
    }
    n++;
    /* fprintf(stderr, "part '%s'\n", token); */
  }
  return n;

 line_clean_up:
  clean_up_parts(*buf, n);
  return 0;
}

int us_interaction(struct uade_command *cmd, int wait_for) {
  struct pollfd ifd;
  int ret, maxlen;
  int i, j;
  int bytes_read;
  char **parts = 0;
  int nparts = 0;

  if (interaction_setup) {
    struct termios tp;
    interaction_setup = 0;
    if (tcgetattr(0, &tp)) {
      perror("uade: can't setup interactive mode");
      goto no_interaction;
    }
    uade_save_terminal();
    tp.c_lflag &= ~(ICANON | ECHO | ECHOE | ECHOK | ECHONL | ECHOPRT);
    if (tcsetattr(0, TCSAFLUSH, &tp)) {
      perror("uade: can't setup interactive mode (tcsetattr())");
      goto no_interaction;
    }
  }

  if (!prompt_shown) {
    printf("uade %d> ", cmd_number);
    fflush(stdout);
    prompt_shown = 1;
    cmd_number++;
  }

  cmd->ret = 0;
  memset(&ifd, 0, sizeof(ifd));

  bytes_read = 0;

  ifd.fd = 0;
  ifd.events = POLLIN;
  ret = poll(&ifd, 1, wait_for ? -1 : 0);
  if (ret > 0) {
    if (ifd.revents & POLLIN) {
      maxlen = sizeof(cmd_buf) - cmd_buf_size;
      if (maxlen > 0) {
	ret = read(ifd.fd, &cmd_buf[cmd_buf_size], maxlen);
	if (ret < 0) {
	  if (errno != EINTR) {
	    goto no_interaction;
	  }
	} else if (ret == 0) {
	  goto no_interaction;
	} else {
	  cmd_buf_size += ret;
	  bytes_read = ret;
	}
      }
    }
  } else if (ret < 0) {
    if (errno != EINTR) {
      goto no_interaction;
    }
  }

  for (i = cmd_buf_size - bytes_read; i < cmd_buf_size;) {
    if (cmd_buf[i] == '\t') {
      break;
    }
    if (cmd_buf[i] == 0x7f) {
      if (i > 0) {
	printf("%c%c%c", 0x1b, 0x5b, 0x44);
	printf(" ");
	printf("%c%c%c", 0x1b, 0x5b, 0x44);
	shift_left(cmd_buf, &cmd_buf_size, i - 1, 2);
	i = i - 1;
      } else {
	shift_left(cmd_buf, &cmd_buf_size, i, 1);
      }
      continue;
    }
    printf("%c", cmd_buf[i]);
    i++;
  }
  if (bytes_read > 0) {
    fflush(stdout);
  }

  ret = scan_cmd_buf();
  if (!ret) {
    tab_state = 0;
    return 0;

  } else if (ret == INTERACTION_EOL) {
    tab_state = 0;

  } else if (ret == INTERACTION_TAB) {
    tab_state++;
    auto_completion();
    return 0;

  } else {
    fprintf(stderr, "uade: odd interaction input\n");
    return 0;
  }

  nparts = divide_line(&parts);

  /* remove the rest of the line from cmd_buf[] */
  i = 0;
  while (i < cmd_buf_size) {
    if (cmd_buf[i] == '\n') {
      i++;
      break;
    }
    i++;
  }
  shift_left(cmd_buf, &cmd_buf_size, 0, i);

  if (!nparts) {
    return 0;
  }

  if (!strcmp("next", parts[0]) || !strcmp("n", parts[0])) {
    cmd->type = UADE_SONG_END;
    goto got_command;
  } else if (!strcmp("m", parts[0])) {
    cmd->type = UADE_REBOOT;
    goto got_command;
  }

  i = 0;
  while (parts[0][i]) {
    if (!isdigit((int) parts[0][i])) {
      goto not_a_number;
    }
    i++;
  }
  cmd->type = UADE_SETSUBSONG;
  cmd->ret = malloc(sizeof(int));
  if (!cmd->ret) {
    goto no_interaction;
  }
  *((int *) cmd->ret) = (int) atoi(parts[0]);
  goto got_command;

 not_a_number:

  if (!strcmp("play", parts[0])) {
    glob_t pglob;
#ifndef GLOB_TILDE
#define GLOB_TILDE 0
#endif
    for (i = 1; i < nparts; i++) {
      if (!glob(parts[i], GLOB_NOCHECK | GLOB_TILDE, 0, &pglob)) {
	for (j = 0;  j < ((int) pglob.gl_pathc); j++) {
	  fprintf(stderr, "enqueuing: %s\n", pglob.gl_pathv[j]);
	  us_enqueue_song(pglob.gl_pathv[j]);
	}
	globfree(&pglob);
      }
    }
    prompt_shown = 0;
    goto no_command;
  }
  
  if (!strcmp("cd", parts[0])) {
    glob_t pglob;
    for (i = 1; i < nparts; i++) {
      if (!glob(parts[i], GLOB_NOCHECK | GLOB_TILDE, 0, &pglob)) {
	for (j = 0; j < ((int) pglob.gl_pathc); j++) {
	  if (chdir(pglob.gl_pathv[j])) {
	    fprintf (stderr, "can't change to directory '%s'\n", pglob.gl_pathv[j]);
	  } else {
	    break;
	  }
	}
	globfree(&pglob);
      }
    }
    prompt_shown = 0;
    goto no_command;
  }
  
  if (!strcmp("ls", parts[0])) {
    char path[PATH_MAX];
    struct dirent **namelist;
    int n;
    char *name;
    if (nparts > 1) {
      strlcpy(path, parts[1], sizeof(path));
    } else {
      strlcpy(path, ".", sizeof(path));
    }
    n = scandir(path, &namelist, 0, alphasort);
    if (n < 0) {
      fprintf(stderr, "can't scan directory '%s'\n", path);
    } else {
      for (i = 0; i < n; i++) {
	name = namelist[i]->d_name;
	if (strcmp(".", name) && strcmp("..", name)) {
	  printf("%s\n", name);
	}
	free(namelist[i]);
      }
      free(namelist);
    }
    prompt_shown = 0;
    goto no_command;
  }

  if (!strcmp("flush", parts[0])) {
    char fname[PATH_MAX];
    while (!us_song_list_empty()) {
      us_get_next(fname, sizeof(fname));
    }
    prompt_shown = 0;
    goto no_command;
  }

  if (!strcmp("help", parts[0]) || !strcmp("h", parts[0])) {
    printf("<number>\tGo to subsong <number>\n");
    printf("cd <dir>\tSwitch to directory <dir>\n");
    printf("f\t\tToggle LED Filter\n");
    printf("flush\t\tFlush playlist\n");
    printf("h, help\t\tDisplay help\n");
    printf("m\t\tNext song\n");
    printf("n\t\tNext (sub)song\n");
    printf("play <pattern>\tEnqueue songs matching shell pattern\n");
    prompt_shown = 0;
    goto no_command;
  }

  if (strlen(parts[0]) == 1) {
    switch (parts[0][0]) {
    case 'f':
      cmd->type = UADE_TOGGLE_LED;
      goto got_command;
    default:
      break;
    }
  }

  printf("unknown command for interaction\n");
  prompt_shown = 0;

 no_command:
  clean_up_parts(parts, nparts);
  return 0;

 got_command:
  clean_up_parts(parts, nparts);
  prompt_shown = 0;
  return 1;

 no_interaction:
  clean_up_parts(parts, nparts);
  cmd->type = UADE_NO_INTERACTION;
  return 1;
}
